/*
 *  <Short Description>
 *  Copyright (C) 2024  Brett Terpstra
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <blt/gfx/window.h>
#include <blt/gfx/renderer/postprocess.h>
#include <blt/gfx/renderer/shaders/postprocess.vert>
#include <blt/gfx/renderer/shaders/pp_screen.frag>
#include <blt/gfx/renderer/shaders/pp_outline_step.frag>
#include <blt/std/ranges.h>

#define __EMSCRIPTEN__

float full_screen_vertices[20] = {
        1.0f, 1.0f, 0.0f, 1.0f, 1.0f,
        1.0f, -1.0f, 0.0f, 1.0f, 0.0f,
        -1.0f, -1.0f, 0.0f, 0.0f, 0.0f,
        -1.0f, 1.0f, 0.0f, 0.0f, 1.0f
};
const unsigned int full_screen_indices[6] = {
        0, 1, 3,
        1, 2, 3
};

template<typename T>
bool once()
{
    static bool ran = true;
    return std::exchange(ran, false);
}

namespace blt::gfx
{
    void pp_engine_t::create()
    {
        {
            vertex_buffer_t vertices_vbo;
            element_buffer_t indices_vbo;
            
            vertices_vbo.create();
            indices_vbo.create();
            
            vertices_vbo.allocate(sizeof(full_screen_vertices), full_screen_vertices);
            indices_vbo.allocate(sizeof(full_screen_indices), full_screen_indices);
            
            screen_vao = std::make_unique<vertex_array_t>();
            const auto tb = screen_vao->bindVBO(vertices_vbo, 0, 3, GL_FLOAT, 5 * sizeof(float), 0);
            screen_vao->bindVBO(tb, 1, 2, GL_FLOAT, 5 * sizeof(float), 3 * sizeof(float));
            screen_vao->bindElement(indices_vbo);
        }
#ifdef __EMSCRIPTEN__
        addStep(std::make_unique<pp_to_screen_step_t>());
#endif
        for (auto& step : steps)
            step->create();
    }
    
    void pp_engine_t::render()
    {
        screen_vao->bind();
        for (const auto& [index, step] : blt::enumerate(steps))
        {
            if (index == 0)
                continue;
            auto& previous = steps[index - 1];
            step->getBuffer().bind(frame_buffer_t::draw_t::BOTH);
            step->draw(*previous);
            if (step->hasShader())
                step->getShader().bind();
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
            glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, 0);
        }
        glBindTexture(GL_TEXTURE_2D, 0);
        frame_buffer_t::unbind();
#ifndef __EMSCRIPTEN__
        steps.back()->getBuffer().blitToScreen(getWindowWidth(), getWindowHeight());
#endif
    }
    
    void pp_engine_t::cleanup()
    {
        screen_vao = nullptr;
        for (auto& v : steps)
            v = nullptr;
        steps.clear();
    }
    
    std::unique_ptr<shader_t> pp_engine_t::createShader(std::string_view fragment)
    {
        auto shader = std::make_unique<shader_t>(shader_postprocess_vert, std::string(fragment));
        shader->bindAttribute(0, "vertex");
        shader->bindAttribute(1, "uv_in");
        return shader;
    }
    
    void pp_engine_t::bind()
    {
        steps.front()->draw();
        steps.front()->getBuffer().bind();
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    }
    
    void pp_to_screen_step_t::create()
    {
        shader = pp_engine_t::createShader(shader_pp_screen_frag);
    }
    
    void pp_to_screen_step_t::draw(pp_step_t& previous)
    {
        glActiveTexture(GL_TEXTURE0);
        previous.getBuffer().getTexture(frame_buffer_t::attachment_t::COLOR0).bind();
    }
    
    void pp_render_target_t::create()
    {
        draw_buffer = frame_buffer_t::make_render_texture(getWindowWidth(), getWindowHeight());
    }
    
    void pp_render_target_t::draw()
    {
        draw_buffer.updateBuffersStorage(getWindowWidth(), getWindowHeight());
    }
    
    void pp_outline_target::create()
    {
        draw_buffer.create();
        draw_buffer.bind();
        
        auto* texture = new texture_gl2D(getWindowWidth(), getWindowHeight(), GL_RGBA8);
        draw_buffer.attachTexture(texture, frame_buffer_t::attachment_t::COLOR0);
        
        auto* mask = new texture_gl2D(getWindowWidth(), getWindowHeight(), GL_RGBA8);
        draw_buffer.attachTexture(mask, frame_buffer_t::attachment_t::COLOR1);
        
        render_buffer_t depth_rbo = render_buffer_t::make_render_buffer(GL_DEPTH24_STENCIL8, getWindowWidth(), getWindowHeight());
        draw_buffer.attachRenderBuffer(depth_rbo, frame_buffer_t::attachment_t::DEPTH_STENCIL);
        
        if (!frame_buffer_t::validate())
            BLT_ERROR("Failed to create render texture framebuffer!");
        frame_buffer_t::unbind();
    }
    
    void pp_outline_target::draw()
    {
        draw_buffer.updateBuffersStorage(getWindowWidth(), getWindowHeight());
    }
    
    void pp_outline_step_t::create()
    {
        draw_buffer = frame_buffer_t::make_render_texture(getWindowHeight(), getWindowHeight());
        shader = pp_engine_t::createShader(shader_pp_outline_step_frag);
        shader->setInt("albedo", 0);
        shader->setInt("mask", 1);
    }
    
    void pp_outline_step_t::draw(pp_step_t& previous)
    {
        draw_buffer.updateBuffersStorage(getWindowWidth(), getWindowHeight());
        glActiveTexture(GL_TEXTURE0);
        previous.getBuffer().getTexture(frame_buffer_t::attachment_t::COLOR0).bind();
        glActiveTexture(GL_TEXTURE1);
        previous.getBuffer().getTexture(frame_buffer_t::attachment_t::COLOR1).bind();
        shader->bind();
        auto v = manager.getScale2D();
        shader->setVec2("viewportSize", static_cast<f32>(getWindowWidth()) * (1.0f/v.x()), static_cast<f32>(getWindowHeight()) * (1.0f / v.y()));
    }
}